/*
 * WBT201.java
 *
 * Created on July 7, 2007, 8:45 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */


package com.pythsoft;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import com.mindprod.ledatastream.LEDataOutputStream;
import java.util.Date;
import java.util.Calendar;
import java.util.TimeZone;
import java.util.Vector;

/**
 *
 * @author J.D. Krijnders
 */

public class WBT201 {
    private static final int LINE_DELIMITER = 10;
    private static final int BT201CHUNK = 4096;
    private static final int RECLEN_WBT201 = 16;
    
    private static final char TRACKMASK = 0x01;
    
    private String name = null;
    private String serialNumber = null;
    private Double hwVersion = null;
    private Double fmtVersion = null;
    private Double swVersion = null;
    private int trackCount = 0;
    private Vector tracks;
    private track currentTrack = null;
    
    private int offset;
    private int lastTrackOffset = 0x0400;
    private char[] lastDateChunk;
    private char[] prevLastDateChunk;
    private Logger logger = null;

    InputStreamReader isr;
    OutputStreamWriter osr;
    boolean loggedIn;
    
    /** Creates a new instance of WBT201 */
    public WBT201(InputStreamReader aIsr, OutputStreamWriter aOsr) {
        isr = aIsr;
        osr = aOsr;
        logger = Logger.getInstance();
        loggedIn = false;
        tracks = new Vector();
        lastDateChunk = new char[4];
        prevLastDateChunk = new char[4];
    }
    
    public boolean login() throws IOException {
        getStringParam("@AL");
        getStringParam("@AL");
        loggedIn = true;
        return true;
    }

    public void logout() throws IOException {
        sendCommand("@AL,2,1");
        loggedIn = false;
    }
    
    public String readLine(char prefix) throws IOException{
        StringBuffer output = new StringBuffer();

        // Read one line and try to parse it.
        int input;
        boolean state = false;
        while ((input = isr.read()) != LINE_DELIMITER) {
            if ((char) input == prefix)
                state = true;
            if (state) {
                output.append((char) input);
                //logger.message(4 , input);
            }
        }
        logger.message(4 , "Last read: " + input);
        
        return output.toString();
    }
    
    public boolean sendCommand(String cmd) throws IOException {
        if (loggedIn)
            while (isr.ready())
                isr.read();
        
        logger.message(4 ,"Going to send: " + cmd);
        //osr.write(cmd + (char) LINE_DELIMITER);
        osr.write(cmd + '\r' + '\n');
        return true;
    }
    public String getStringParam(String cmd) throws IOException {
        sendCommand(cmd);
        return readString(cmd);
    }
    
    public String readString(String cmd) throws IOException {
        boolean loop = true;
        int i = 0;
        String ret = new String();
        while (loop){
            if (i++ > 10)
                loop = false;
            ret = readLine('@');
            logger.message(4 ,"readLine returned: " + ret);
            if (ret.startsWith(cmd)) {
                loop = false;
            }  else {
                sendCommand(cmd);
            }
        }
        ret = ret.substring(cmd.length());
        if (ret.startsWith(","))
            ret = ret.substring(1);
        
        logger.message(4 ,"Device returned in readString: " + ret);
        return ret;
    }
    
    public Integer getIntParam(String cmd) throws IOException{
        String s = getStringParam(cmd);
        Integer ret = new Integer(Integer.parseInt(s));
        return ret;
    }
    
    public Double getDoubleParam(String cmd) throws IOException {
        Double d;
        try {
            d = Double.valueOf(getStringParam(cmd));
        } catch (NumberFormatException e)
        {
            d = new Double(0.0);
            e.printStackTrace();
        }
        return d;
    }
    
    private char readWithTimeout(int timeout) throws IOException
    {
        char ret = '\0';
        if (isr.ready()) {
            ret = (char) isr.read();
        } else {
            try {
                int i;
                for (i = 0; i < 10; i++) {
                    Thread.sleep(timeout/10);
                    if (isr.ready())
                    {
                        ret = (char) isr.read();
                        break;
                    }  
                }
                if (i == 10)
                    throw new IOException();
            } catch(InterruptedException e) {
            
            }
        }
        return ret;
    }
    
    public char[] getData(String cmd, int length) throws IOException {
        char ret[] = new char[length]; //length]; // = new char(length);
        sendCommand(cmd);
        try {
            int j;
                for (j = 0 ;j < 25; j++)
                    if (!isr.ready())
                        Thread.sleep(100);
                if (!isr.ready())
                    throw new IOException();
                
                int readBytes = isr.read(ret, 0, length);
        } catch(InterruptedException e) {
            }
        return ret;
    }
    
    public String calcChecksum(char[] data, int len) {
        int checksum = 0;
        int i = 0;
        for (i = 0; i < len; i++)
            checksum ^= data[i];
        String CS = new String(Integer.toHexString(checksum));
        if (CS.length() == 1) // Fix zero padding
            CS = "0" + CS;
        return CS.toUpperCase(); // WBT201 returns uppercase checksums;
    }
    
    public int countTracks(char[] data, int len) {
        int i, count = 0;
        
        for (i = 0; i < len; i = i + 16) {
            boolean track = (data[i] & TRACKMASK) != 0x00;
            if ( track ) {
                // If no current track, don't save it
                if ( currentTrack != null) {
                    // Update track time and store track in vector
                    currentTrack.setStopDate(chunkDate(swapBytesToInt(prevLastDateChunk, 0)));
                    tracks.addElement(currentTrack);
                }
                currentTrack = new track(trackCount + count );
                currentTrack.setStartDate(chunkDate(swapBytesToInt(data, i + 2)));
                currentTrack.setNumberOfPoints(0);
                currentTrack.setOffset(offset + i + 1024);
                currentTrack.setDistance(0);
                currentTrack.setTime(0);
                count++;
            }

            int latitude = swapBytesToInt(data, i + 6);
            int longitude = swapBytesToInt(data, i + 10);
            currentTrack.addPoint((double)latitude / 10000000d, (double)longitude / 10000000d);
            prevLastDateChunk = lastDateChunk;
            lastDateChunk[0] = data[i + 2];
            lastDateChunk[1] = data[i + 3];
            lastDateChunk[2] = data[i + 4];
            lastDateChunk[3] = data[i + 5];
        }
        return count;
    }
    
    private int swapBytesToInt(char[] data, int i) {
            int accum = 0;
            int shiftBy = 0;
            for (shiftBy = 0; shiftBy < 32; shiftBy += 8) {
                accum |= (data[i] & 0xff ) << shiftBy;
                i++;
            }
            return accum;
    }
    
    public Date chunkDate(int data) {
        Calendar date = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
        date.setTime(new Date());
        date.set(Calendar.SECOND, ((data) >> 0) & 0x3F);
        date.set(Calendar.MINUTE, ((data) >> 6) & 0x3F);
        date.set(Calendar.HOUR_OF_DAY, ((data) >> 12) & 0x1F);
        date.set(Calendar.YEAR, 2000 + (((data) >> 26) & 0x3F));
        date.set(Calendar.MONTH, (((data) >> 22) & 0x0F) - 1);
        date.set(Calendar.DATE, ((data) >> 17) & 0x1F);
        return date.getTime();
        
    }
    
    public String getName() throws IOException {
        if (name == null) {
            name = getStringParam("@AL,7,1");
        }
        return name;
    }
    
    public Double getHwVersion() throws IOException {
        if (hwVersion == null)
            hwVersion = getDoubleParam("@AL,8,1");
        return hwVersion;
    }
    
    public Double getSwVersion() throws IOException {
        if (swVersion == null)
            swVersion = getDoubleParam("@AL,8,2");
        logger.message(4 ,"Got software version: " + swVersion);
        return swVersion;
    }
    
    public Double getFmtVersion() throws IOException {
        if (fmtVersion == null)
            fmtVersion = getDoubleParam("@AL,8,3");
        return fmtVersion;
    }
    
    public String getSerialNumber() throws IOException {
        if (serialNumber == null)
            serialNumber = getStringParam("@AL,7,3");
        return serialNumber;
    }
    
    public int bytesToBeRead() throws IOException {
        Integer logStart = getIntParam("@AL,5,1");
        Integer logStop = getIntParam("@AL,5,2");
        return logStop.intValue() - logStart.intValue(); 
    }
    
    public int getCapacity() throws IOException {
        Integer addrStart = getIntParam("@AL,5,9");
        Integer addrStop = getIntParam("@AL,5,10");
        return addrStop.intValue() - addrStart.intValue();
    }
    
    public int getTrackCount() {
        if (trackCount == 0)
            logger.message(0, "Asked for track count before reading data");
        
        return trackCount;
    }
    
    public int getPointCount() throws IOException {
        return bytesToBeRead()/16;
    }
    
    public void readAndWrite(LEDataOutputStream fos, org.netbeans.microedition.lcdui.WaitScreen screen) throws IOException {
        Integer logStart = getIntParam("@AL,5,1");
        Integer logStop = getIntParam("@AL,5,2");
        long startms = new Date().getTime();
        long diffms;
        trackCount = 0;
        logger.message(4 ,"Reading " + logStart + " to " + logStop);
        char[] data = null;
        int want = logStop.intValue() - logStart.intValue();
        int totalWant = want;
        loggedIn = true;
        
        int numReads = want / BT201CHUNK;
        logger.message(4 ,"Going to read " + Integer.toString(numReads) + " blocks");
        offset = logStart.intValue();
        String CScalc;
        String CSphone;
        int retry = 0;
        while (want > BT201CHUNK) {
            data = getData("@AL,5,3," + Integer.toString(offset), BT201CHUNK);
            if (data.length == BT201CHUNK) {
                CScalc = calcChecksum(data, BT201CHUNK);
                String tmp = readString("@AL,CS");
                CSphone = tmp.substring(0,2);
                logger.message(4 ,CScalc + "<=>" + CSphone);
                if (CSphone.equalsIgnoreCase(CScalc))  { //  || CSphone.equalsIgnoreCase("0" + CScalc)
                    fos.write(new String(data).getBytes());
                    trackCount += countTracks(data, BT201CHUNK);
                    int percentage = (int) (1000f * offset / totalWant);
                    diffms = new Date().getTime() - startms;
                    long speed = (long) 10000f * offset / (1024 * diffms);
                    screen.setText(// "Read " + offset + "/" + totalWant + " bytes\n" +
                            percentage/10f +"% @ " + (speed/10f) + " kb/s");
                    want = want - BT201CHUNK;
                    offset = offset + BT201CHUNK;
                    retry = 0;
                }
            } else {
                retry = retry + 1;
                logger.message(0 ,"Retry # " + retry);
                if (retry > 11) {
                    logger.message(0 ,"Giving up");
                    break;
                }
            }
        }
        logger.message(0,"Reading remainder");
        data = getData("@AL,5,3," + Integer.toString(offset), want);
        CScalc = calcChecksum(data, want);
        String tmp = readString("@AL,CS");
        CSphone = tmp.substring(0,2);
        logger.message(4 ,CScalc + "<=>" + CSphone);
        if (CSphone.equalsIgnoreCase(CScalc)) {
            logger.message(4 ,"Checksums matched");
            logger.message(4 ,"Copyping from " + 0 + " to offset, " + offset + ", length " + want);
            fos.write(new String(data).getBytes());
            trackCount += countTracks(data, want);
        }
        tracks.addElement(currentTrack);
        tracks.addElement(new track(trackCount, offset + want + 1024, 1));
        trackCount++;
        fos.writeShort(0x07);
        fos.write(new String(data).substring(data.length-14).getBytes());
        int k;
        for (k = 0; k < trackCount; k++) {
            track tr = (track) tracks.elementAt(k);
            tr.writeTrack(fos);
        }
        logger.message(0, "Found " + trackCount + " tracks");
    }
    
    public void erase() throws IOException {
        sendCommand("@AL,5,6");
    }
    
    public int getConfigDynamicPlatform() throws IOException {
        return 1; //this.getIntParam("@AL");
    }
    
    public Vector getTracks() {
        if (trackCount == 0)
            logger.message(0, "Asked for track count before reading data");

        return tracks;
    }
}
